import re
from typing import Any

from .base import ValidationError
from .string import StringValidator
from ..types import RawCustomField


class EmailValidator(StringValidator):
    EMAIL_REGEX = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"

    @classmethod
    def __is_good_email(cls, value: str) -> bool:
        return bool(re.fullmatch(cls.EMAIL_REGEX, value))

    @classmethod
    def validate(
        cls,
        value: Any,
        raw_field: RawCustomField
    ) -> None:
        super().validate(value, raw_field)

        if not cls.__is_good_email(value):
            raise ValidationError(f"Must be an email! Your value: {value}")



import os
from requests import Response

from .base import BaseApi
from ..utils.decorators import resp_to_json
from ..utils.exceptions import AuthTokenMissing


class ClickUpApi(BaseApi):
    """Class that proxy the ClickUp api."""
    
    endpoint = "https://api.clickup.com/api/v2"
    

    def _perform_auth(self) -> None:
        """Attaches ClickUp API token to API headers."""

        token = os.environ.get("CLICKUP_AUTH_TOKEN")

        if not token:
            raise AuthTokenMissing(
                "Please, set `CLICKUP_AUTH_TOKEN` in your environment."
            )

        self.headers.update({
            "Authorization": token
        })

    @resp_to_json
    def get_task(self, task_id: str) -> Response:
        return self.get(f"{self.endpoint}/task/{task_id}")

    @resp_to_json
    def get_batch_tasks(self, list_id: str, page: int=0) -> Response:
        return self.get(
            f"{self.endpoint}/list/{list_id}/task",
            params=dict(page=page)
        )

    @resp_to_json
    def create_task(self, list_id: str, **kwargs) -> Response:
        return self.post(f"{self.endpoint}/list/{list_id}/task", json=kwargs)

    @resp_to_json
    def update_task(self, task_id: str, **kwargs) -> Response:
        return self.put(f"{self.endpoint}/task/{task_id}", json=kwargs)

    @resp_to_json
    def get_custom_fields(self, list_id: str) -> Response:
        return self.get(f"{self.endpoint}/list/{list_id}/field")

    @resp_to_json
    def set_custom_field(self, task_id: str, field_id: str, **kwargs) -> Response:
        return self.post(f"{self.endpoint}/task/{task_id}/field/{field_id}", json=kwargs)




from typing import Any

from .base import BaseValidator, ValidationError
from ..types import RawCustomField
from ...list import ClickUpList


class RelationValidator(BaseValidator):
    """
    Validates that only specific ClickUpList can be set to task, e.g.
    
    class ChildList(ClickUpList):
        name = TextField(field_name="Some text")

    class ParentList(ClickUpList):
        child = RelationField(ChildList, field_name="My child")

    >> parent = ParentList()
    >> child = ChildList()
    >> parent.child = child  # GOOD
    >> parent.child = "Test" # Error
    """

    @classmethod
    def _is_correct_relation_class(
        cls, 
        value: ClickUpList, 
        raw_field: RawCustomField
    ) -> bool:
        """
        Method that will check if value, that is wanted to be set is
        correct RelationField list class

        The idea behind this is that list_relationship type_config must
        have `subcategory_id` value that represents the id of related list.
        Related list value can also be grabbed from LIST_ID variable in 
        ClickUpList classes.

        So, summing up, if 

        raw_field.type_config.subcategory_id == value.__class__.LIST_ID
        
        so the value is good.
        """
        value_list_id = str(value.__class__.LIST_ID)
        list_id_from_field = raw_field["type_config"]["subcategory_id"]

        return value_list_id == list_id_from_field 

    @classmethod
    def validate(
        cls,
        value: Any,
        raw_field: RawCustomField
    ) -> None:
        super().validate(value, raw_field)
        
        if not isinstance(value, ClickUpList):
            raise ValidationError(
                "Value must be object of type/subtype ClickUpList"
            )

        value_class = value.__class__.__name__

        if not cls._is_correct_relation_class(value, raw_field):         
            raise ValidationError(
                f"You can't set {value_class} for {raw_field['name']} field "
            )

        if not value.id:
            raise ValidationError(
                f"{value_class} must have `id`. Please, create your object."
            )



from .base import CustomField
from ..utils.validators import ForbidValuesValidator


class FormulaField(CustomField):
    """
    Formula fields allow you ONLY to get value. They will set automatically.
    """
    TYPE = "formula"
    VALIDATOR = ForbidValuesValidator




from datetime import date, datetime
from typing import Any

from .base import BaseValidator, ValidationError
from ..types import RawCustomField


class DateValidator(BaseValidator):
    """Validates that `value` is `datetime` or `date` value."""

    @classmethod
    def validate(cls, value: Any, raw_field: RawCustomField) -> None:
        super().validate(value, raw_field)

        if not (type(value) in [datetime, date]):
            raise ValidationError(
                f"{raw_field['name']} must be `datetime` or `date`."
            )



from typing import Any

from .base import BaseValidator, ValidationError
from ..types import RawCustomField


class NumberValidator(BaseValidator):
    """Validates that value is `float` or `int`."""

    @classmethod
    def validate(cls, value: Any, raw_field: RawCustomField) -> None:
        if type(value) not in [int, float]:
            raise ValidationError("Must be `int` or `float` value")

class PositiveNumberValidator(NumberValidator):
    """Validates that value is `float` or `int` that is >= 0."""

    @classmethod
    def validate(cls, value: Any, raw_field: RawCustomField) -> None:
        super().validate(value, raw_field)

        if value < 0:
            raise ValidationError("Must be number that is >= 0")



from .base import CustomField
from ..utils.types import RawCustomField
from ..utils.validators import CheckboxValidator


class CheckboxField(CustomField):
    TYPE = "checkbox"
    VALIDATOR = CheckboxValidator
    
    def get_value(self, raw_field: RawCustomField) -> bool:
        value = raw_field.get("value")

        if not value:
            return False

        return True if value == "true" else False 

    def set_value(self, value: bool, raw_field: RawCustomField) -> None:
        if value:
            raw_field["value"] = "true"
        else:
            raw_field["value"] = "false"




from typing import Optional

from .base import CustomField
from ..services import geo_grabber
from ..utils.types import RawCustomField



class LocationField(CustomField):
    TYPE = "location"

    def set_value(self, value: str, raw_field: RawCustomField) -> None:
        location = geo_grabber.search(value)
        raw_field['value'] = location.to_dict()

    def get_value(self, raw_field: RawCustomField) -> Optional[str]:
        value = raw_field.get("value")
        
        if value is None:
            return

        return value['formatted_address']




from dataclasses import dataclass
from geopy.geocoders import Nominatim
from geopy.exc import GeopyError


class AddressNotFound(Exception):
    """Raised when address is not found via geopy."""

@dataclass
class Location:
    latitude: float
    longitude: float
    address: str

    def to_dict(self) -> dict:
        """Return same dict structure as ClickUp provide for Location Field"""
        return dict(
            location=dict(lat=self.latitude,
                          lng=self.longitude),
            formatted_address=self.address
        )

class GeolocationGrabber:
    """
    Service that will find out the details about specific address,
    using geopy library.
    """

    def __init__(self):
        self._geolocator = Nominatim(user_agent=self.__class__.__name__)

    def search(self, address: str) -> Location:
        try:
            raw_location = self._geolocator.geocode(address)
        except GeopyError:
            raise GeopyError(
                "Something happen with Geopy request."
                "Please, check that address is valid."
            )

        if not raw_location:
            raise AddressNotFound(f"`{address}` is not valid or not found.") 

        return Location(
            address=address,
            latitude=raw_location.latitude,
            longitude=raw_location.longitude
        )



from typing import Any, TypedDict, List


class PostCustomField(TypedDict):
    id: str
    value: str

class CustomFieldsCache:
    """
    Class that will store all new custom_fields to perform custom fields
    update later.

    It will store the values after we set correct value for that, e.g.
    ```
    MyClickUpList.attr = <New Value>
    ```
    """

    def __init__(self):
        self._storage = [] 

    def add(self, field_id: str, value: Any):
        self._storage.append(PostCustomField(
            id=field_id,
            value=value
        ))

    def get(self) -> List[PostCustomField]:
        result = self._storage[:]
        self.clear()
        return result

    def clear(self):
        self._storage.clear()


class ClientListsRegistry:
    """
    Simple cache storage that will store all of the client's 
    lists names and classes at the beginning of the runtime.
    """
    _storage = {}
    
    @classmethod
    def get(cls, list_name: str):
        return cls._storage.get(list_name, None)

    @classmethod
    def set(cls, list_name: str, value):
        cls._storage[list_name] = value

    @classmethod
    def update(cls, clients_lists: dict):
        cls._storage.update(clients_lists)



from requests import Response
from typing import Callable
from .exceptions import BadRequest


def resp_to_json(func: Callable) -> Callable:
    """
    Check if the return value is `Response` and check if
    status code is 2xx. If it is not 2xx, BadRequest will be raises.
    """
    def wrapper(*args, **kwargs) -> dict:
        resp = func(*args, **kwargs)

        if not isinstance(resp, Response):
            return resp
        
        if resp.ok:
            return resp.json()

        raise BadRequest(f"Status {resp.status_code}\nContent: {resp.json()}") 
    return wrapper



from typing import Any

from .base import ValidationError
from .string import StringValidator 
from ..types import RawCustomField


class UrlValidator(StringValidator):
    """Validates that string is URL."""
    
    @classmethod
    def validate(
        cls,
        value: Any,
        raw_field: RawCustomField
    ) -> None:
        super().validate(value, raw_field)

        if not value.startswith('http'):
            raise ValidationError(f"Must be a good url! Your value: {value}")



from typing import Any

from .base import BaseValidator, ValidationError
from ..types import RawCustomField


class CheckboxValidator(BaseValidator):
    """Validates that only `True`/`False` can be set."""

    @classmethod
    def validate(cls, value: Any, raw_field: RawCustomField) -> None:
        super().validate(value, raw_field)

        if not (value is True or value is False):
            raise ValidationError(
                f"{raw_field['name']} must be `True` or `False`."
            )



from abc import abstractmethod, ABC
from typing import Any

from ..types import RawCustomField


class ValidationError(Exception):
    """Raised by Validator when condition is not matched."""


class BaseValidator(ABC):
    """Interface for future Validators."""

    @classmethod
    @abstractmethod
    def validate(cls, value: Any, raw_field: RawCustomField) -> None:
        ...

class ForbidValuesValidator(BaseValidator):
    """Will not allow to set any values."""

    @classmethod
    def validate(cls, value: Any, raw_field: RawCustomField) -> None:
        if value:
            raise ValidationError(
                "You are not allowed to set any value for that field!"
            )



from typing import List, Union

from .base import CustomField
from ..utils.types import RawCustomField
from ..utils.validators import LabelsValidator


class LabelsField(CustomField):
    TYPE = "labels"
    VALIDATOR = LabelsValidator

    def get_value(self, raw_field: RawCustomField) -> List[str]:
        value_ids = raw_field.get('value')

        if value_ids is None:
            return []
 
        options = raw_field['type_config']['options']

        def get_name_by_id(id: str) -> str:
            return next(filter(
                lambda option: option['id'] == id, options
            ))['label']

        return list(map(get_name_by_id, value_ids))

    def set_value(
        self, value: Union[str, List[str]], raw_field: RawCustomField
    ) -> None:
         
        options = raw_field['type_config']['options']
        valid_options = list(filter(
            lambda option: option['label'] in value, options
        ))

        previous_values = raw_field.get('value') or []

        raw_field['value'] = [ 
            *previous_values,
            *map(lambda option: option['id'], valid_options)
        ]



from typing import Any, TypedDict, List, Optional


class RawCustomField(TypedDict):
    id: str
    type: str
    name: str
    type_config: dict
    value: Optional[Any]

class RawTask(TypedDict):
    id: str
    description: str
    url: str
    name: str
    type: str
    custom_fields: List[RawCustomField]

class FieldToUpdate(TypedDict):
    id: str
    value: Any



from datetime import datetime, timedelta
from typing import Optional

from .base import CustomField
from ..utils.validators import DateValidator
from ..utils.types import RawCustomField


class DateField(CustomField):
    TYPE = "date"
    VALIDATOR = DateValidator

    @classmethod
    def __clickup_timestamp_to_datetime(cls, timestamp: str) -> datetime:
        tmp = float(timestamp) / 1000
        return datetime.fromtimestamp(tmp)

    @classmethod
    def __datetime_to_clickup_timestamp(cls, date: datetime) -> str:
        return str(datetime.timestamp(date) * 1000) 

    def get_value(self, raw_field: RawCustomField) -> Optional[datetime]:
        timestamp = raw_field.get("value")

        if timestamp is None:
            return

        return self.__clickup_timestamp_to_datetime(timestamp)       

    def set_value(self, value: datetime, raw_field: RawCustomField) -> None:
        raw_field['value'] = self.__datetime_to_clickup_timestamp(value)




class Singleton:
    """Class that implements singletone pattern."""
    
    _instances: dict = dict()

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances: 
            instance = super().__new__(cls, *args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls] 



from typing import Any, Optional, Type

from ..list import ClickUpList
from ..utils.types import RawCustomField
from ..utils.validators import StringValidator


class CustomFieldNotFound(Exception):
    """Raised if raw custom field not found in raw clickup list."""

class CustomFieldIsNull(Exception):
    """Raised when None is set to required field."""

class CustomField:
    """
    Descriptor class that will set the workflow with ClickUp task's custom
    fields data.

    This workflow is based on 2 main ideas:
    1) __get__ method MUST return simple humanreadable value
    2) __set__ method MUST get simple humanreadable value that will be
       validated by CustomField VALIDATOR

    It will give a possibility to easily work with custom fields.

    This class is a base CustomField. There are also many others subclasses
    that represent different ClickUp custom fields (e.g. TextField, 
    NumberField, RelationField, DropDownField, etc.)

    There is one important pre-get/set workflow. Before working with attribute, 
    following steps will run on our `instance` (class, that will use CustomField):
    1) Check that `instance` is `ClickUpList`
    2) Check that `CustomField.field_name` == `instance.name`  
    3) Check that `CustomField.TYPE` == `instance.type`
    """
    TYPE = None
    VALIDATOR = StringValidator

    def __init__(self, field_name: str, required: bool=False):
        self.field_name = field_name
        self.required = required

    def __get__(self, instance: Any, instance_type: Type) -> Any:
        """
        Descriptor GET method that will:
        1) Do pre-get workflow (check class docstring)
        2) Run getter method (get_value)
        """
        raw_field = self._raw_field(instance)
        return self.get_value(raw_field)

    def __set__(self, instance: ClickUpList, value: Any) -> None:
        """
        Descriptor SET method that will:
        1) Check if there is a value and if the field is not
           required we can set None.
        2) Do pre-set workflow (check class docstring)
        3) Validate value to be the right type
        4) Run setter method (set_value)
        5) Store new value in cache
        """

        raw_field = self._raw_field(instance)

        if value is None:
            if self.required:
                raise CustomFieldIsNull(f"`{self.field_name}` can not be None")

            raw_field['value'] = None
        else:
            self.__validate_value(value, raw_field)
            self.set_value(value, raw_field)

        self.__store_value_in_cache(instance, raw_field)

    def get_value(self, raw_field: RawCustomField) -> Optional[Any]:
        """
        Getter that will return humanreadable value from raw dict-like 
        custom field

        By default it will return raw custom field's `value` attr.
        Can be implemented in subclasses
        """
        return raw_field.get('value', None)

    def set_value(
        self, value: Any, raw_field: RawCustomField
    ) -> None:
        """
        Setter that will set correct value based on humanreadable `value`.

        By default it will set `value` to the raw custom field's `value` attr.
        Can be implemented in subclasses
        """
        raw_field['value'] = value


    def __validate_value(
            self,
            value: Any,
            raw_field: RawCustomField
        ) -> None:
        """
        Method that will invoke specific validation for current class.
        This method is used BEFORE setting the attribute.
        """

        self.VALIDATOR.validate(value, raw_field)

    def _raw_field(self, instance: Any) -> RawCustomField:
        """
        Pre-get/set workflow implementation.
        Do following steps:
        1) Check that `instance` is `ClickUpList`
        2) Check that `CustomField.field_name` == `instance.name`  
        3) Check that `CustomField.TYPE` == `instance.type`
        4) Return good custom field or raise error
        """
        instance = self.__check_instance_type(instance) 
        field = self.__get_field_by_name_and_type(instance)

        if not field:
            raise CustomFieldNotFound(
                "There is no such field with\n"
                f"- type: {self.TYPE}\n"
                f"- name: {self.field_name}"
            )

        return field

    def __check_instance_type(self, instance: Any) -> ClickUpList:
        if not isinstance(instance, ClickUpList):
            raise TypeError("Instance must be `ClickUpList`")
        return instance

    def __get_field_by_name_and_type(
        self, instance: ClickUpList
    ) -> Optional[RawCustomField]:
        return next(filter(
            lambda field: 
            field['name'] == self.field_name and  
            field['type'] == self.TYPE,
            instance.custom_fields
        ), None)

    def __store_value_in_cache(
        self,
        instance: ClickUpList,
        raw_field: RawCustomField
    ) -> None:
        id = raw_field['id']
        value = raw_field.get('value')
 
        if not value: return
        
        if self.TYPE == 'list_relationship':
            only_ids = list(map(lambda item: item['id'], value))
            value = dict(add=only_ids)
        instance._fields_cache.add(field_id=id, value=value)





from .text import TextField
from .short_text import ShortTextField
from .url import UrlField
from .email import EmailField
from .drop_down import DropDownField
from .relation import RelationField
from .phone import PhoneField
from .checkbox import CheckboxField
from .number import NumberField
from .currency import CurrencyField
from .location import LocationField
from .date import DateField
from .labels import LabelsField
from .formula import FormulaField


__all__ = [
    "TextField",
    "ShortTextField",
    "UrlField",
    "EmailField",
    "DropDownField",
    "RelationField",
    "PhoneField",
    "CheckboxField",
    "NumberField",
    "CurrencyField",
    "LocationField",
    "DateField",
    "LabelsField",
    "FormulaField"
]



from typing import Any, List

from .base import ValidationError
from .string import StringValidator
from ..types import RawCustomField


class DropDownValidator(StringValidator):
    """
    Validates that value is the correct `name` variable from
    field's type_config options.
    """

    @classmethod
    def __has_good_option(cls, value: str, options: List[dict]) -> bool:
        good_option = next(filter(
            lambda option: option['name'] == value,
            options
        ), None)
        return good_option is not None
                
    @classmethod
    def validate(
        cls,
        value: Any,
        raw_field: RawCustomField
    ) -> None:
        super().validate(value, raw_field)

        options = raw_field['type_config']['options']

        if not cls.__has_good_option(value, options):
            valid_names = list(map(lambda option: option['name'], options))
            raise ValidationError(
                "Invalid option!\n"
                f"Your value: {value}\n"
                f"Possible values: {valid_names}"
            )



from typing import Type, List, Union 
from .base import CustomField
from ..list import ClickUpList
from ..services import clickup_api
from ..services.cache import ClientListsRegistry
from ..utils.types import RawCustomField
from ..utils.validators import RelationValidator


class ClientListNotFound(Exception):
    """Raised when client list class not found."""

class RelationField(CustomField):
    TYPE = "list_relationship"
    VALIDATOR = RelationValidator

    def __init__(self, related_list: Union[str, Type[ClickUpList]], field_name: str):
        super().__init__(field_name)
        self._related_list = related_list
        
    @property
    def related_list(self):
        """
        Property that will grab correct client list from registry on runtime.
        """
        related_list = self._related_list

        if isinstance(related_list, str):
            related_list_name = related_list
            related_list = ClientListsRegistry.get(related_list)
            if not related_list:
                raise ClientListNotFound(
                    f"There is no such class: {related_list_name}"
                )

        return related_list

    def set_value(
        self, 
        value: ClickUpList,
        raw_field: RawCustomField
    ) -> None:
        current_values = raw_field.get("value")

        if not current_values:
            current_values = []

        if current_values:
            existed_ids = list(map(lambda item: item['id'], current_values))
            if value.id in existed_ids:
                return

        raw_field['value'] = [*current_values, dict(id=value.id)]

    def get_value(self, raw_field: RawCustomField) -> List[ClickUpList]:
        values = raw_field.get("value")

        if values is None:
            return []

        def build_correct_task(value: dict) -> ClickUpList:
            raw_task = clickup_api.get_task(task_id=value['id'])
            return self.related_list(raw_task)

        return list(map(
            lambda value:
            build_correct_task(value),
            values
        ))



from .checkbox import CheckboxValidator
from .drop_down import DropDownValidator
from .email import EmailValidator
from .number import NumberValidator, PositiveNumberValidator
from .relation import RelationValidator
from .string import StringValidator, NumericStringValidator
from .url import UrlValidator
from .uuid import UuidValidator
from .date import DateValidator
from .labels import LabelsValidator
from .base import ForbidValuesValidator



from requests import Session

from ..utils.datastructures import Singleton


SAFE_METHODS = ("get", "head", "option")
VALID_METHODS = (*SAFE_METHODS, "post", "put", "patch", "delete")

class BaseApi(Session, Singleton): 
    """
    Base API class that will be used for different 3rd party APIs.

    Each subclass API will be created once (using Singleton) and can
    be used globally.
    """

    endpoint = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs) 
        self._perform_auth()

    def _perform_auth(self):
        """
        Interface for API authentication.
        Must be implemented in subclasses.
        """




from typing import Optional

from .base import CustomField
from ..utils.types import RawCustomField
from ..utils.validators import DropDownValidator


class DropDownField(CustomField):
    TYPE = 'drop_down'
    VALIDATOR = DropDownValidator

    def get_value(self, raw_field: RawCustomField) -> Optional[str]:
        orderindex = raw_field.get('value')

        if orderindex is None:
            return

        options = raw_field['type_config']['options']
        good_option = next(filter(
            lambda option: option['orderindex'] == orderindex,
            options
        ))

        return good_option['name']

    def set_value(self, value: str, raw_field: RawCustomField) -> None:
        """
        value is the `name` variable of type_config >> options.
        """
        options = raw_field['type_config']['options']

        good_option = next(filter(
            lambda option:
            option['name'] == value,
            options
        ))
        
        raw_field['value'] = good_option['orderindex']




import re
from typing import Any

from .base import ValidationError
from .string import StringValidator
from ..types import RawCustomField


class UuidValidator(StringValidator):
    """Validates that string is UUID."""

    UUID_REGEX = r"^[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[0-9a-f]{4}-[0-9a-f]{12}$"

    @classmethod
    def __is_good_uuid(cls, value: str) -> bool:
        return bool(re.fullmatch(cls.UUID_REGEX, value))

    @classmethod
    def validate(
        cls,
        value: Any,
        raw_field: RawCustomField
    ) -> None:
        super().validate(value, raw_field)

        if not cls.__is_good_uuid(value):
            raise ValidationError(
                f"Must be a UUID string! Your value:  {value}"
            )



from __future__ import annotations
from functools import cached_property
from typing import Any, List, Union, Optional, Generator

from .services import clickup_api
from .services.cache import CustomFieldsCache, ClientListsRegistry
from .utils.types import RawTask, RawCustomField
 

class TaskIdNotFound(Exception):
    """Raised, when there is no `id` found in task object."""

class TaskFromWrongList(Exception):
    """
    Raised, when we fetch task by id from correct space, but for wrong list
    """

class ReadOnlyTaskField(Exception):
    """
    Raised, when there was a try of manual update of task's `id` field.
    """

class ClientListsLookup(type):
    """
    Metaclass that will store each client class to the registry.
    So, when a client define a class
    ```
    class ClientsClass(ClickUpList):
        ...fields
    ```
    the `ClientsClass` will be stored in the registry as 
    {
        client_list_name: client_list_class
    }
    """
    
    def __new__(cls, name, bases, attrs):
        cls = type.__new__(cls, name, bases, attrs)

        if bases: 
            ClientListsRegistry.update({
                name: cls 
                for base in bases
                if base == ClickUpList
            })

        return cls

class ClickUpList(metaclass=ClientListsLookup):
    """
    Base ClickUp list representation, that will have default ClickUp task
    properties. 

    This list has CRUD functionality that use ClickUpApi.

    This class is the base class for client ClickUp lists. E.g.
    ``` 
    class ClientClass(ClickUpList):
        field_1 = CustomField(field_name="Client Field Name")
    ``` 

    It will give the possibility to easily work with custom ClickUp lists.
    """
    LIST_ID = None

    def __init__(self, raw_task: Union[RawTask, dict]=dict()):
        self._raw_task = raw_task
        self._fields_cache = CustomFieldsCache()

    @property
    def id(self) -> Optional[str]:
        return self._raw_task.get("id")

    @id.setter
    def id(self, value: Any) -> None:
        raise ReadOnlyTaskField(
            "You can't set new `id` manually."
            "`id` will be added to the task when you will get/create a task" 
        )

    @property
    def name(self) -> Optional[str]:
        return self._raw_task.get("name")

    @name.setter
    def name(self, value: Any) -> None:
        if not isinstance(value, str):
            raise TypeError("`name` attribute must be str")
        self._raw_task['name'] = value

    @property
    def description(self) -> Optional[str]:
        return self._raw_task.get("description")

    @description.setter
    def description(self, value: Any) -> None:
        if not isinstance(value, str):
            raise TypeError("`description` attribute must be str")
        self._raw_task['description'] = value

    @property
    def url(self) -> Optional[str]:
        return self._raw_task.get("url")

    @url.setter
    def url(self, value: Any) -> None:
        raise ReadOnlyTaskField(
            "You can't set new `url` manually."
            "`url` will be added to the task when you will get/create a task" 
        )

    @cached_property
    def custom_fields(self) -> List[RawCustomField]:
        """
        Try to get custom fields from `raw_task`. If they are not found,
        they will be pulled from clickup.
        """
        custom_fields = self._raw_task.get('custom_fields')

        if custom_fields:
            return custom_fields

        return self.__get_custom_fields_from_api()

    def __get_custom_fields_from_api(self) -> List[RawCustomField]:
        raw_payload = clickup_api.get_custom_fields(self.LIST_ID)
        return raw_payload['fields']

    # CRUD functionality
    @classmethod
    def _get_all_tasks_as_chunks(cls) -> Generator:
        """
        Method that will fetch all tasks by chunks of 100 tasks.
        There is a generator as a return value to avoid extra api requests
        for cases where we need to find some of the tasks, e.g. when we
        invoke `get_by_name` method it wants us to run through all of the 
        tasks and find out with correct `name` value, so better solution
        here is to fetch tasks as chunks of 100 tasks and look through the 
        chunk, to find out the good name. If name is found - stor intration.
        """

        page = 0
        tasks = clickup_api.get_batch_tasks(cls.LIST_ID, page).get('tasks')

        yield tasks if tasks else []

        page += 1
    
    @classmethod
    def get_by_id(cls, task_id: str):
        """Class method, that will fetch task by its id."""
        raw_task = clickup_api.get_task(task_id=task_id)

        if raw_task is None:
            return

        if raw_task['list']['id'] != cls.LIST_ID:
            raise TaskFromWrongList(
                "You have tried to fetch task from wrong list."
                "Please, check your's object list class."
            )

        return cls(raw_task)

    @classmethod
    def get_by_name(cls, task_name: str):
        """Class method, that will fetch task by its name."""
        for tasks in cls._get_all_tasks_as_chunks():
            raw_task = next(filter(
                lambda task: task['name'] == task_name, tasks 
            ), None)

            if raw_task:
                return cls(raw_task)

    def create(self):
        """
        Method, that will get current raw_task 
        values and push them to ClickUp.
        """
        custom_fields = self._fields_cache.get()
        body = dict(
            name=self.name,
            custom_fields=custom_fields
        )
        raw_task = clickup_api.create_task(self.LIST_ID, **body)
        self._raw_task = raw_task

    def update(self):
        """
        Method, that will update existed ClickUp task.
        It will
        1) Check task_id. If there is no id, so the task is not existe.
        2) Update base task field (Only name by now)
        3) Update custom fields
        """
        if not self.id:
            raise TaskIdNotFound("Can not update unexisted Task")

        body = dict(
            name=self.name,
            description=self.description
        )

        self._raw_task = clickup_api.update_task(self.id, **body)
 
        custom_fields = self._fields_cache.get()

        if custom_fields:
            for field in custom_fields:
                clickup_api.set_custom_field(
                    task_id=self.id,
                    field_id=field['id'],
                    value=field['value']
                )



from typing import Any

from .base import BaseValidator, ValidationError
from ..types import RawCustomField


class StringValidator(BaseValidator):
    """Validates that value is `str` type."""
    
    @classmethod
    def validate(cls, value: Any, raw_field: RawCustomField) -> None:
        if not isinstance(value, str):
            raise ValidationError("Must be string")

class NumericStringValidator(StringValidator):
    """Validate that value is `str` type but only with numbers [0-9]."""

    @classmethod
    def validate(cls, value: str, raw_field: RawCustomField) -> None:
        super().validate(value, raw_field)

        if not value.isnumeric():
            raise ValidationError("Must be string with numbers [0-9]")








from typing import Any, List, Union

from .base import BaseValidator, ValidationError
from ..types import RawCustomField


class OptionsListValidator(BaseValidator):
    """
    Validates that value is one of the `options` in `type_config`.

    Valid setter options:
    - task.attr = "valid_option_as_str"
    - task.attr = ["valid_option_as_str_1", "valid_option_as_str_1"]
    """
    NAME_KEY = None

    @classmethod
    def __has_good_options(
        cls, value: Union[str, List[str]], valid_options_names: List[str]
    ) -> bool:
        return all([name in valid_options_names for name in value])
                
    @classmethod
    def validate(cls, value: Any, raw_field: RawCustomField) -> None:
        
        if not (isinstance(value, list) or isinstance(value,  str)):
            raise ValidationError("Must be `str` or `list`")

        if isinstance(value, str):
            value = [value]

        options = raw_field['type_config']['options']
        valid_names = list(map(lambda option: option[cls.NAME_KEY], options))

        if not cls.__has_good_options(value, valid_names):
            raise ValidationError(
                "Invalid option(s)!\n"
                f"Your value(s): {value}\n"
                f"Possible values: {valid_names}"
            )



from typing import Optional
from .base import CustomField
from ..utils.types import RawCustomField
from ..utils.validators import NumberValidator


class NumberField(CustomField):
    TYPE = "number"
    VALIDATOR = NumberValidator

    def get_value(self, raw_field: RawCustomField) -> Optional[float]:
        value =  raw_field.get("value")
        
        if not value:
            return

        return float(value)
